<?php

namespace Reloadify\Service;

use Doctrine\DBAL\Connection;
use Reloadify\Entity\AbandonedCart\AbandonedCartDefinition;
use Reloadify\Entity\AbandonedCart\AbandonedCartEntity;
use Shopware\Core\Checkout\Cart\Cart;
use Shopware\Core\Checkout\Cart\Error\ErrorCollection;
use Shopware\Core\Defaults;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\Doctrine\RetryableQuery;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\Uuid\Uuid;

class AbandonedCartService
{
    private Connection         $connection;
    protected EntityRepository $abandonedCartRepository;

    public function __construct(Connection $connection, EntityRepository $abandonedCartRepository)
    {
        $this->connection = $connection;
        $this->abandonedCartRepository = $abandonedCartRepository;
    }

    /**
     * @param string  $id
     * @param Context $context
     * @return AbandonedCartEntity|null
     */
    public function getAbandonedCartWithId(string $id, Context $context)
    {
        $criteria = new Criteria([$id]);
        /** @var AbandonedCartEntity */
        return $this->abandonedCartRepository->search($criteria, $context)->first();
    }

    /**
     * @param string  $token
     * @param Context $context
     * @return AbandonedCartEntity|null
     */
    public function getAbandonedCartWithToken(string $token, Context $context)
    {
        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::TOKEN_FIELD, $token));

        /** @var AbandonedCartEntity */
        return $this->abandonedCartRepository->search($criteria, $context)->first();
    }

    public function setAbandonedCartToken(string $id, string $token, Context $context)
    {
        $this->abandonedCartRepository->update(
            [
                [
                    'id'                                 => $id,
                    AbandonedCartDefinition::TOKEN_FIELD => $token,
                ],
            ],
            $context
        );
    }

    /**
     * @param Context $context
     * @return void
     */
    public function deleteOldAbandonedCarts(Context $context)
    {
        $criteria = new Criteria();
        $date = new \DateTime();
        $date->sub(new \DateInterval('P30D'));
        $criteria->addFilter(
            new RangeFilter(
                'createdAt',
                [RangeFilter::LTE => $date->format(Defaults::STORAGE_DATE_TIME_FORMAT)]
            )
        );

        $carts = $this->abandonedCartRepository->searchIds($criteria, $context);
        $ids = \array_map(
            static function ($id) {
                return ['id' => $id];
            },
            $carts->getIds()
        );
        $this->abandonedCartRepository->delete(
            $ids,
            $context
        );
    }

    /**
     * @param string  $token
     * @param Context $context
     * @return void
     */
    public function deleteAbandonedCart(string $token, Context $context)
    {
        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::TOKEN_FIELD, $token));

        $abandonedCartId = $this->abandonedCartRepository->searchIds($criteria, $context)->firstId();

        if (empty($abandonedCartId)) {
            return;
        }

        $this->abandonedCartRepository->delete(
            [
                [
                    'id' => $abandonedCartId,
                ],
            ],
            $context
        );
    }

    public function deleteAbandonedCartByCustomerId(string $customerId, Context $context)
    {
        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::CUSTOMER_ID_FIELD, $customerId));

        $abandonedCartId = $this->abandonedCartRepository->searchIds($criteria, $context)->firstId();

        if (empty($abandonedCartId)) {
            return;
        }

        $this->abandonedCartRepository->delete(
            [
                [
                    'id' => $abandonedCartId,
                ],
            ],
            $context
        );
    }

    /**
     * @param Cart    $cart
     * @param string  $token
     * @param string  $customerId
     * @param Context $context
     * @return void
     * @throws \Doctrine\DBAL\ConnectionException
     */
    public function saveAbandonedCart(Cart $cart, string $token, string $customerId, string $salesChannelId, string $languageId, string $currencyId, float $totalPrice, float $netPrice, Context $context): void
    {
        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::TOKEN_FIELD, $token));
        $payload = $cart->getLineItems()->getElements();

        $abandonedCartId = $this->abandonedCartRepository->searchIds($criteria, $context)->firstId();
        if (empty($abandonedCartId)) {
            $abandonedCartId = Uuid::randomBytes();
        }
        else {
            $abandonedCartId = Uuid::fromHexToBytes($abandonedCartId);
        }
        $this->connection->beginTransaction();

        $sql = <<<'SQL'
        INSERT INTO reloadify_abandoned_cart (id, token, cart, payload, customer_id, sales_channel_id, language_id, currency_id, amount_total, amount_net, created_at)
        VALUES (:id, :token, :cart, :payload, :customer_id, :sales_channel_id, :language_id, :currency_id, :amount_total, :amount_net, :now)
        ON DUPLICATE KEY UPDATE
            id = :id,
            token = :token,
            cart = :cart,
            payload = :payload,
            customer_id = :customer_id,
            sales_channel_id = :sales_channel_id,
            language_id = :language_id,
            currency_id = :currency_id,
            amount_total = :amount_total,
            amount_net = :amount_net,
            updated_at = :now;
        SQL;

        $serializedCart = $this->serializeCart($cart);

        $data = [
            'id'          => $abandonedCartId,
            'token'       => $token,
            'cart'        => $serializedCart,
            'payload'     => json_encode($payload),
            'customer_id' => Uuid::fromHexToBytes($customerId),
            'sales_channel_id' => Uuid::fromHexToBytes($salesChannelId),
            'language_id' => Uuid::fromHexToBytes($languageId),
            'currency_id' => Uuid::fromHexToBytes($currencyId),
            'amount_total' => $totalPrice,
            'amount_net' => $netPrice,
            'now' => (new \DateTime())->format(Defaults::STORAGE_DATE_TIME_FORMAT),
        ];

        try {
            $query = new RetryableQuery($this->connection, $this->connection->prepare($sql));

            $query->execute($data);
        }
        catch (\Throwable $e) {
            //Do nothing
        }

        $this->connection->commit();
    }

    private function serializeCart(Cart $cart): string
    {
        $errors = $cart->getErrors();
        $data = $cart->getData();

        $cart->setErrors(new ErrorCollection());
        $cart->setData(null);

        $serializedCart = serialize($cart);

        $cart->setErrors($errors);
        $cart->setData($data);

        return $serializedCart;
    }

}
